/* minimization of M(natural)-convex function
 *  This is an implementation of greedy algorithm
 *  (improved steepest descent method).
 *  The calculation cost is O(dimension) per step.
 *  cf. K. Murota: Discrete convex analysis (Japanese), 2001, p.230.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "odicon.h"

/* returns index (for plus) if improvment was found. */
/* returns -1 if no improvment was found. */
static int best_improve(int dim, double func(int,int*), int x[],
                        int minus, int natural) {
    int i, min_i = -1;
    double f, min_f;

    iteration++;
    min_f = func(dim, x);  oracle_call++;
    x[minus]--;

    for (i=0; i<dim+natural; i++) {
        if (i == minus) continue;
        x[i]++;
        f = func(dim, x);  oracle_call++;
        if (f < min_f) {  min_f = f;  min_i = i;  improve_call++;  }
        x[i]--;
    }

    x[minus]++;
    return min_i;
}


/* returns the minimum value of func(x)    */
/* INPUT  func   : pointer to the function */
/* INPUT  init[] : the initial solution    */
/* OUTPUT init[] : the minimized solution  */
static double minimize(int dim, double func(int,int*), int init[],
                       int natural) {
    int *x, plus, minus;

    counter_reset();
    if (!is_feasible(dim, func, init)) return PLUS_INFINITY;
    status_print(dim, func, init);

    assert(natural == 0 || natural == 1);
    x = ivector_new(dim + natural);
    ivector_copy(dim, x, init);

    for (minus=0; minus<dim+natural; minus++) {
        while ((plus = best_improve(dim, func, x, minus, natural)) >= 0) {
            x[plus]++;
            x[minus]--;
        }
        loop_print(dim, func, x);
    }
    ivector_copy(dim, init, x);
    free(x);

    counter_print();
    minimizer_print(dim, func, init);
    return func(dim, init);
}


double mconv_minimize2(int dim, double func(int,int*), int init[]) {
    return minimize(dim, func, init, 0);
}

double mgconv_minimize2(int dim, double func(int,int*), int init[]) {
    return minimize(dim, func, init, 1);
}
